#ifndef RAND_UNRANK_H
#define RAND_UNRANK_H

#include <sys/sem.h>	
#include <sys/ipc.h>
#include <semaphore.h>	//POSIX semaphores. 

/*

    prand.so - A block parallelization of glibc's random() function
    Copyright (C) 2003  Jason Main (jmain@onyx.boisestate.edu)
						Lewis Hall

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* *************** BLOCK PARALLELIZATION OF GLIB's random() FUNCTION *************

CONTACT INFORMATION: If you are having problems with the following code first read the
  comments in this file. If the problems persist, please contact:

            Jason Main
            jmain@onyx.boisestate.edu

The student concerned with time may wish to skip to the 'FUNCTION PROTOTYPES' section,
or the 'COMPILING AND LINKING' section

WARNING: This file assumes that random.c has not changed. As the code has a copyright 
from 1988, this will probally be the case for some time but make sure that you are 
useing a glibc with a random.c no newer than December 18th, 2003. Very bad things
can happen if random.c changes without this file being updated.

***************************************************************************
CURRENT STATUS / REV
***************************************************************************

Revision: 1.0 December 18th, 2003

Notes: This library currently supports the unranking of states within glibc's
       random() function. Unranking is the act of taking a number and consistently
	   mapping that number to a state. In this case the unrank function maps
	   a number to a state so that random() acts as a nextState() function. This
	   is used to support a block parallelization of glibc's random() function.
	   Several helper functions are provided to simplify the use of this function
	   in a parallel programming setting. Technically unranking is a term that
	   denotes imposing an absolute ordering on states. This unrank function 
	   instead imposes a relative ordering (relative to the current state)
	   which for purposes of using random() in a parallel program is actually
	   more useful.

Future Direction / Progress:

rankRand(): This function is in theoretical development. I currently know how to
		rank (invert the unrankRand() functions actions to go from a state to a
		number) on the linear congruence type, but am awaiting further incite to
		be able to rank a lagged fibonacci generator. Currently the problem is
		half solved for lagged fibonacci generators.
Better running order: Current runtime is O( polySize^3 * log(stride) ). I am 
		perusing means to reduce this to O( polySize^2  + log(stride)*polySize ).
		Theoretical work still needs to be done.
Increasing the range of sizes supported by glibc: This would be implemental in
		about an hour except that without a better runtime, unrankRand()s overhead
		may become prohibitive to its use. Right now there is a good match between
		the running time of unrankRand() and the polynomial sizes that random() 
		supports.
More optimized unrankRand(): The ideas are ready to optimise unrankRand() more 
		thoroughly with its current algorithm. This will be done if the 
		order type of unrankRand() is not improved first.
Better srandom() function: Currently srandom() does not permit the user to span
		the entire state space of a lagged fibonacci generator. Because of this
		the seeding is less than ideal. I am in the process of fixing this.
ready on the randomObject stuct: Currently this paramiter is implimented as an 
		int which incurs a 1 or 2 clock cycle hit for wait_random(). This is
		much smaller then a semaphore would give (a semaphore isnt needed because
		only one the value is never reset) but it can be done better with function
		pointers and macros. The added speed of course comes at the price that a 
		corrupt randomObject would probally cause a seg-fault.
free unrankRand from any dependance on globalState: currently only one thread can
		call unrankRand at a time because of the need to keep global state information.
		a call to unrankRand() is automaticly protected from reenterancey by a semaphore.
		The restriction of keeping unrankRand() from being called simultaniously by two
		threads will be removed next revision.
***************************************************************************

This library implements:
    int* createProcessRandomTable() - creates random tables handy for sorting routines, etc
    int seedParallel() - Sets up random() without performing any calls to random()
    unsigned long long seedParallelTotal() - similar to seedParallel() but it divides the range
                                         for you.
    int unrankRand() - Takes a stride (negative or positive) and moves the state of random()
                   to duplicate the action of repeatedly calling random(). All functions
                   are based off of this one.

And if you're doing something tricky the following functions:
    par_initialize() - Initialize the thread safe version. There is the small possibility of a memory
                   leak if you don't call this function but the thread interface will work
                   without it. There is also the chance of a mis-syncronization.
				   This function is not needed if you dont use the par_* interface
    par_generateRandomObject() - create a random number stream. All of the other par_
                        functions use this parameter to alter their values. If NULL is
                        passed into any of the par_ function for the stream field then
                        they operate on glibc's random state.
    par_setsate()     - Thread safe version of the corresponding glibc function
    par_initstate()     - Thread safe version of the corresponding glibc function
    par_srandom()     - Thread safe version of the corresponding glibc function
    par_unrankRand() - Thread safe version of unrank()
    par_random()     - Not recommended. Thread safe, but slow. If you're not calling
                   random() while your calling any of the setstate(), etc functions
                   this is not needed. See wait_random() instead.
    wait_random()     - This function will block until par_unrank() is called. After this
                   has occurred there is no time penalty for using wait_random() vs.
                   random().
    random_rollback() - Function to help figure out what numbers were received from random
                        if you did not use the thread safe interface and you called
                        random() from one thread at the same time that unrankRand() was
                        being called from another
    par_getSem()     - These function get and take the semaphore that protects the par_ interface
                   when a user has this semaphore he/she knows that any of the par_ functions
                   will block. This does not prevent the none par_ functions from executing.
                   NOTE: The wait_random()
    par_releaseSem()    - This function will release the semaphore received by par_getSem()


This library is meant to implement a block parallelization of glibc's standard random
function. Once the random()'s internal state has been set by the functions of this 
file, the regular random() function can be called to generate random() numbers. The 
idea is to break up the numbers so that that the union of all random numbers used by
the process is the same as it would be for 1 process generating a contiguous sequence 
of random numbers. Say that letters represent the return values from successive calls 
to random(). Then the situation we wish to duplicate is:

    The serial process calls random() 50 times receiving the random numbers....
SERIAL: abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWX

    Each process calls random() 10 times receiving the values...
PROCESS0: abcdefghij
PROCESS1: klmnopqrst
PROCESS2: uvwxyzABCD
PROCESS3: EFGHIJKLMN
PROCESS4: OPQRSTUVWX

Leapfrog parallelization is not currently supported. 
Neither is independent sequence parallelization.

The principle function used for this is called unrankRand(). unrankRand() permutes the state so
that each process effectively starts with its random() function so that the numbers it 
generates correspond to its block of random numbers. It takes a parameter called stride
that represents how many random() calls from the current state you want to simulate. In other
words the following code snippets are functionally equivalent (although unrankRand() is faster):

    ITERATIVE                
for(i=0;i<1000000;i++)
    random();

    USING UNRANK
unrankRand(1000000);


To use this in a parallel program we would use the following format:

SERIAL:
srandom(SEED);    //Consume the whole range of random numbers
for    (i=0;i<n;i++) {
    tmp = random();
    ...

PARALLEL:        //Each process uses a fraction of the total range...
srandom(SEED)
unrankRand( myProcessID * (n/numberOfProcessors) );
for    (i=0;i < (n/numberOfProcessors);i++) {
    tmp = random();
    ...

Some care must be taken if n does not evenly divide by the number of processors. Functions
are provided that simplify this. In addition unrankRandom() works on the number of randoms
consumed not the number of loop iterations so if each process consumes say 2 random numbers
for every loop, then the call to unrank must be adjusted so that n is 2 times the number of
loop iterations

NOTE: unrankRandom()'s runtime is typically less than 10 milliseconds for a gigahertz machine
or better. For problems that take less than a second to solve this may be significant 
overhead. In this case you can set up random() with a linear congruent generator with 
something like:

    initstate(SEED, malloc(8), 8);    //Force random() to use Linear congruent
    unrankRand(wherever);

unrankRand() for linear congruent is very quick (on the order of 2 microseconds for 
gigahertz machines). The bigger question would be 'What are you doing that you have
to parallelize something that only takes a second?'

*/


/*        NOTES:

Parallel Interface:
 NOTE: glibc's random() function uses a linear congruent generator (TYPE_0) if less than
        32 bytes of information is available for storing the state and uses a lagged fibonacci
        generator (See Knuth AOP Vol 2...) otherwise. There are 4 supported sizes of LFG's
        supported ranging from 32 bytes of information (TYPE_1) to 256 bytes (TYPE_4).
        These functions assume that the processes have the same sized generator
        setup. If no call to setstate() or initstate() is used then
        the r.n.g. will be a TYPE_3 by default. If you call either
        of these function you must make sure the processes are initialized
        to the same r.n.g. type. For instance unrankRand() will not 
        behave as designed if process 1 calls:
        
        initstate(123456, myArray_process1, 256);    //let random() use 256 bytes
        seedParallel(123456, me,10000000);

        and process 2 calls:

        initstate(123456, myArray_process2, 64);        //Give random() 64 bytes
        seedParallel(123456, me,10000000);

        The polynomial type can be guaranteed by either calling initstate()
        with the same number as the third parameter, or more simply by
        just using the defaults and never calling initstate().

NOTE:   Because of glibc's implementation the last random number generated does not 
        fully encode the state (like it does with linear congruence). This has a
        few implications. 
        1) A process may receive the same random number from random() without it 
            necessarily being the case that the sequence has entered a loop. 
            In other words the period may (and probably will) be larger then the 
            first repeated random() number. For instance with a TYPE_3 random number
            generator (glibc's default), most initial states have a period of size
            2^31(2^31 - 1) or roughly a 62 bit number. Successive calls to random()
            must repeat a number within (2^32 + 1) calls because there are only 
            2^32 possible return values with a 32 bit machine word.
        2) For exactly the same reason as 1), two differing processes may get the same 
            random number from random without it being the case that the processes will
            then start repeating all numbers (I.E. correlated streams of random numbers)
		3) State space is much larger for a lagged fibonacci generator then the cycle length.
		    For instance the default TYPE_3 generator has a 2^31(2^31 - 1) sized period.
			It however effectively has 31 words or 992 bits of state information. Because
			of this calling srandom(INIT_1) and srandom(INIT_2) where INIT_1 != INIT_2 will
			most likely not result in the two states being part of the same sequence of random 
			numbers but instead will most likely result in two completely independent 
			sequences. In other words no number of calls to random() will make the two
			sync up.

COMPILING AND LINKING:

Building the shared library:
This is a sample make file for building the shared library:

	CC=gcc
	all: libprand.so
	
	libprand.so:	rand.o
		$(LD) -shared -o $@ $?

	clean:
		rm --force *.o libprand.so a.out core rand.o


Use it to build a shared library named libprand.so out of rand.h and rand.c.
Your system (and PVM running on it) needs to know the run time location of the shared 
library, so add the following two lines to your .bashrc file:

	LD_LIBRARY_PATH=$LD_LIBRARY_PATH:<absolute path to where libprand.so is located>
	export LD_LIBRARY_PATH
 

This will make it so the run time shared library loader can find libprand.so.
Using the shared library in your code:

Include rand.h.

When compiling your code, link in the shared library by adding the following to the 
compile command:

	-L<location of libprand.so> -lprand

*/

/*    FUNCTION PROTOTYPES
 The following functions simplify the common usage of unrankRand() as a
 means of distributing random numbers across a parallel cluster. They assume
 there is only one thread calling random() related functions up until the time
 that the initialization is complete
*/

#define RANDOM_SEED   0xF3C3347D	//The function createProcessRandomTable() always uses this
									//seed...
		
int* createProcessRandomTable(int me, int numProc, unsigned long n,
                            unsigned long* myRange);
/*
PURPOSE:    This function creates a table of random values for each process. This action
            is useful for testing programs like parallel search routines or parallel
            sorting algorithms, where a set of integers is distributed across a cluster.
			All processes are started off with a TYPE_3 random number generator initialized
			initialized consistently with a serial program that starts by calling:
			srandom(RANDOM_SEED);

PARAMITERS:
 me:        The process ID of the process calling this function (ID range from 0 to numProc-1)
 numProc:   The number of processes being used in the parallelization (often called p)
 n:         The number of random integers to generate. This many numbers are distributed
              evenly across the cluster with each process receiving about n/numProc of them
 myRange:   A pointer where the function will store how many integers the caller was allocated
              A process will either be allocated n/myRange or n/myRange+1 numbers so that the 
              total range across the cluster is equal to n. If myRange is 0 it is ignored.

RETURN VALUE:
 The function returns an array of size myRange where random integers have been stored. At
 this point the process has been initialized with a subsequence and is ready for work. The
 function returns NULL on an error. Check the errno value against the error codes listed later
 in this file if this occurs. Note for very small problem sizes it may occur that some
 process has no work to do. In this case myRange is set to 0 and the function returns a
 invalid but non-null pointer (equal to -1).

USAGE:

Typical usage goes something like:

    SERIAL (single processor version used for comparison purposes)
array = createProcessRandomTable(0,1,n,0);  //array[] now contains n random numbers
mySort(array,n);                            //Serial sort...

    PARALLEL
array = createProcessRandomTable(myProc,p,n,&range); //array[] now contains range random numbers
myParallelSort(array,range);                        //Sort the numbers...

*/

int seedParallel(unsigned long seed, int me, unsigned long long int numsPerProc);
/*
PURPOSE: This function initializes a processes random number generator. Each process will be
         set up so that it can call random() numsPerProc times before it's stream of 
         random numbers coincides with another processes. 

         To use this function each process would have to know:
            1) Its own process number (where IDs start at 0 and run up continuously to p-1)
            2) How many numbers will be consumed by each processor. For this
               function to work all processes must call random() an equal number of times
            3) A random seed that is the same value at every process. This can be a
                pound define or can be part of the information that is sent to each process

PARAMITERS:
 seed: A random seed given to all processes
 me:   An int that represents this process. This function assumes that
       the process IDs start at 0 and increase contiguously
 numsPerProc:   The number of random numbers that will be consumed by one
                process. This function uses this to tell where to initialize
                each process. Each process must call random() the same number of 
				times for this function to work.

RETURN VALUE:
 The function returns false if the action failed. Error codes are listed farther down
 in this file.

USAGE: 
 This function is typically used when it would be wasteful to allocate space for
 all randoms. For instance consider this code for estimating Pi with Monte Carlo 
 integration:

PARALLEL
 nums = iterPerProc*2;	//Consume 2 randoms per loop iteration
 seedParallel(MY_SEED,me,nums);		//Seed
 for( i =0,total = 0; i< ittrPerProc; i++)
 {
	x= random();	//Find a random point in the (x,y) plain
	y= random();

    square = x*x + y*y;	//Calculate the square of the distance from (0,0)

	if(square < RAND_MAX)
		total++;
 }
 // 4*(total/iterPerProc) is now a good estimate of pi. We send this partial result
 // to the master process to incorporate the numbers from each process...
 ....

SERIAL
 srandom(MY_SEED);	
 //seedParallel(MY_SEED,0,0); // This would work also for seeding
 for( i =0,total = 0; i< n; i++)	//n = ittrPerProc*numProc from above
 {
	x= random();	//Find a random point in the (x,y) plain
	y= random();

    square = x*x + y*y;	//Calculate the square of the distance from (0,0)

	if(square < RAND_MAX)
		total++;
 }
 Pi = 4.0 * ((double)total) / ((double)iterPerProc);
 ....

*/

long long seedParallelTotal(unsigned long seed, 
                                     int me,
                                     int p,
                                     long long int itr,
                                     long long int grain);
/*
PURPOSE: This function is like seedParallel() except that it splits 
         the number range for the user.

PARAMITERS:
itr: The total number of loop iterations. The total number of randoms
     consumed will be itr*grain
me: My process ID where process ID's range from 0 to p-1
p:  number of processes
seed: a random seed that must be equal on all processes
grain: Granularity of the usage of random numbers. When using block
       parallelization, the number of random numbers used per
       main loop iteration must be taken into account. For instance
       if 2 randoms are used per loop iteration then it makes no
       sense to give a process an odd number of random numbers. grain
       allows the function to know how to doll out work

RETURN VALUE:
The function returns the number of loop iterations to perform.
The function returns a negative number upon an error and sets the
errno value appropriately.

USAGE: 
Picking up the code for estimating pi from seedRandom() the following
code would allow the total number of loop iterations to not evenly 
divide the number of processors.

 iterPerProc = seedParallelTotal(MY_SEED,		//All slaves use the same seed 
                                  me,			//I am me. Who else would I be
                                  nProc,		//Number of processors
                                  n,			//We were asked to use n iterations
												//(i.e. n points in (x,y) space) over
												//the entire cluster
                                  2);			//Two numbers per process
 for( i =0,total = 0; i< ittrPerProc; i++)
 {
	x= random();	//Find a random point in the (x,y) plain
	y= random();

    square = x*x + y*y;	//Calculate the square of the distance from (0,0)

	if(square < RAND_MAX)
		total++;
 }
 // 4*(total/iterPerProc) is now a good estimate of pi. We send this partial result
 // to the master process to incorporate the numbers from each process...
 ....	//We must send total and iterPerProc now

SERIAL	 Same program as shown for seedParallel except numProc need not divide n
 srandom(MY_SEED);	
 //seedParallelTotal(MY_SEED,0,1,n,2); // This would work also for seeding
 for( i =0,total = 0; i< n; i++)
 {
	x= random();	//Find a random point in the (x,y) plain
	y= random();

    square = x*x + y*y;	//Calculate the square of the distance from (0,0)

	if(square < RAND_MAX)
		total++;
 }
 Pi = 4.0 * ((double)total) / ((double)iterPerProc);
 ....

*/
                                        
int unrankRand(long long int stride);
/*
PURPOSE:
This function increments random()'s
state in a manner equivalent to stride may calls to random().
Negative numbers are allowed in which case unrankRand() restores
the state to what it would have been -stride many iterations 
previous.

PARAMITERS:
 stride: A number representing how many calls to random() you want to emulate
         if stride is positive the this line:
			while(stride--) random();
		 is functionally equivalent to this line:
			unrankRandom(stride);
		  
		 If stride is passed as a negative number then these lines:
			tmp = -stride;
			while(tmp--) random();
			unrank(stride);
		 Form a really fancy no-op and leave random()s state ultimately unaffected

RETURN VALUE:
The function returns true if there was an error. See below for potential error codes

USAGE:
 NOTE: To use this function properly you must initialize random()
      to a known state (say with srandom() setstate()...
      see the man pages for details of these functions.

 Say you whished to parallelize searching a range of randoms for a specific number 
 (a birthday perhaps). This could be done with:

PARALLEL:
 srandom(SEED);
 unrank(numsPerProc*me);	//Here n must be divisible by the number of processors
 for(i=0;i<numsPerProc;i++)	//see the seedRandom(), etc functions for code that handles
							//this for you automatically
	if(random() != SPECIAL_NUMBER);
		break;
 ...

SERIAL:
 srandom(SEED);
 for(i=0;i<n;i++)
	if(random() != SPECIAL_NUMBER);
		break;
 ...
*/


/*
 Wrapper functions:
 These functions wrap the standard glibc functions for setting state.
 Call these functions when opposing threads may be calling unrankRand()
 and a function like setstate(), srandom(), or initstate(). IMHO you
 should ask yourself why are you trying to do this before you worry about
 whether there is an interface to allow it, but its your code. If you
 don't care what happens the glibc functions are still available.
*/

/*
    par_setsate()     - Thread safe version of the corresponding glibc function
    par_initstate()     - Thread safe version of the corresponding glibc function

    par_unrankRand() - Thread safe version of unrank()
    par_random()     - Not recommended. Thread safe, but slow. If you're not calling
                   random() while your calling any of the setstate(), etc functions
                   this is not needed. See wait_random() instead.
    wait_random()     - This function will block until par_unrank() is called. After this
                   has occurred there is no time penalty for using wait_random() vs.
                   random().
    random_rollback() - Function to help figure out what numbers were received from random
                        if you did not use the thread safe interface and you called
                        random() from one thread at the same time that unrankRand() was
                        being called from another
*/


struct randomObject{	//This structure holds information to allow allow multiple random
	long int* state;	//streams to exist within the code
	long int *fptr;
	long int *rptr;
	long int *end_ptr;
	int rand_type;
	int rand_deg;
	int rand_sep;

	int ready;

	sem_t sem;	//Semaphore protection object. A lot like the Mafia. They will protect
				//you but ya gotta pay...
};

struct randomObject* par_generateRandomObject(int type,unsigned long seed);
int par_freeRandomObject(struct randomObject* ro);
int par_srandom(unsigned int seed, struct randomObject* ro);
int par_getSem(struct randomObject* ro);
int par_releaseSem(struct randomObject* ro);
long int par_random_raw(struct randomObject* ro);
int par_initialize();
long int wait_random(struct randomObject* ro);
//char* par_exportRandomObject(struct randomObject* ro);

/*
PURPOSE:
PARAMITERS:
RETURN VALUE:
USAGE:
*/

/*
ERROR RETURN VALUES FOR unrankRand(), and corresponding functions
Fatal errors are giving a negative value
*/
#define UNRANK_RAND_FAILED_MEMORY_ALLOC        -1
    //Memory could not be allocated to perform the requested action
#define UNRANK_RAND_EXPORT_SIZE_TOO_SMALL    -2
#define UNRANK_UNRECOGNIZED_RAND_TYPE        -3
#define UNRANK_RAND_EXPORT_POINTER_MISMATCH  -4
    //Between import and export of random()s state the state array
    //has changed address. Probably not what is wanted
    //In the microsecond that it took to change states someone has
    //called the 
#define INVALID_NUMBER_OF_PROCS				 -5
	//An numProc parameter was passed with an invalid range. The number
	//of processes must be a strictly positive number
#define UNRANK_FAILED_TO_GET_SEMAPHORE		-6
	//The functions to handle semaphores failed
#define UNRANK_ASKED_TO_RELEASE_A_NONE_EXISTANT_SEMAPHORE	-7
	//Somone tried to release somthing that wasnt there
#define UNRANK_BAD_POINTER			-8
	//Usually a 0 pointer was passed where it was not allowed

//Non-fatal errors have positive error codes
#define UNRANK_RAND_EXPORT_SIZE_MISMATCH    1
#define UNRANK_RAND_IMP_EXP_POINTER_MISMATCH    2
#define UNRANK_RAND_EXPORT_RAND_CALLED_SINCE_LAST_IMPORT 3

//#define PARALLEL_RAND_DEBUG        //Debug flag for testing the code (define for testing)
                                        //Compiling with this flag exports the internal state
                                        //information of unrankRand()


#ifndef PARALLEL_RAND_DEBUG   //Do we export variables....
#define RANDOM_DEC  static
#else
#define RANDOM_DEC  
#endif

#ifdef PARALLEL_RAND_DEBUG    //Export some internal information
extern int degrees[];
extern int seps[];
extern long int randtbl[];
extern long int *fptr;
extern long int *rptr;
extern long int *state;
extern int rand_type;
extern int rand_deg;
extern int rand_sep;
extern long int *end_ptr;

extern const unsigned long invPowerArray[31];
int exportSlibcState();
int importSlibcState();
void identity(unsigned long** a,int n);
int initStaticArray(unsigned long*** p,int n);
void deallocStaticArray(unsigned long*** p);
unsigned long invPowerOfA(unsigned long power);
unsigned long fastPower(unsigned long a, unsigned long n);
unsigned long fastPoly(unsigned long a, unsigned long n);
unsigned long fastPolyItr(unsigned long* powers, unsigned long x,unsigned long n);
int log_2(unsigned long n);
void matrixMult(unsigned long** left,unsigned long** right,
                unsigned long** res,int n);
void updateStateMatrix(unsigned long ** a, int n);
void createPermMatrix(unsigned long** a);
void createPermMatrixRev(unsigned long** a);
void iterativeRand(long stride);
long int t_random_backwords();    //Needed for moving through internal 
long int t_random();            //state space
#endif      //#ifdef PARALLEL_RAND_DEBUG

#endif        //#ifndef RAND_UNRANK_H

